/*
* Copyright(c) 1997-2001 Id Software, Inc.
* Copyright(c) 2002 The Quakeforge Project.
* Copyright(c) 2006 Quetoo.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or(at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
* 
* See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <ctype.h>

#include "gl_local.h"

void R_Clear(void);

viddef_t viddef;

refimport_t ri;

int QGL_TEXTURE0, QGL_TEXTURE1;

model_t *r_worldmodel;

float gldepthmin, gldepthmax;

glconfig_t gl_config;
glstate_t gl_state;

image_t *r_notexture;  // use for bad textures
image_t *r_particletexture;  // little dot for particles

entity_t *currententity;
model_t *currentmodel;

cplane_t frustum[4];

int r_visframecount;  // bumped when going to a new PVS
int r_framecount;  // used for dlight push checking

int c_brush_polys, c_alias_polys;

float v_blend[4];  // final blending color
byte color_white[4] = {255, 255, 255, 255};
byte color_black[4] = {0, 0, 0, 128};

void GL_Strings_f(void);

// view origin
vec3_t vup;
vec3_t vpn;
vec3_t vright;
vec3_t r_origin;

float r_world_matrix[16];
float r_base_world_matrix[16];

// screen size info
refdef_t r_newrefdef;

int r_viewcluster, r_viewcluster2, r_oldviewcluster, r_oldviewcluster2;

cvar_t *r_norefresh;
cvar_t *r_drawentities;
cvar_t *r_drawworld;
cvar_t *r_speeds;
cvar_t *r_fullbright;
cvar_t *r_novis;
cvar_t *r_nocull;
cvar_t *r_lerpmodels;

cvar_t *r_lightlevel;

cvar_t *gl_nosubimage;
cvar_t *gl_allow_software;

cvar_t *gl_vertex_arrays;

cvar_t *gl_ext_palettedtexture;
cvar_t *gl_ext_multitexture;
cvar_t *gl_ext_compiled_vertex_array;

cvar_t *gl_log;
cvar_t *gl_bitdepth;
cvar_t *gl_drawbuffer;
cvar_t *gl_driver;
cvar_t *gl_lightmap;
cvar_t *gl_modulate;
cvar_t *gl_nobind;
cvar_t *gl_round_down;
cvar_t *gl_picmip;
cvar_t *gl_skymip;
cvar_t *gl_wireframe;
cvar_t *gl_ztrick;
cvar_t *gl_finish;
cvar_t *gl_clear;
cvar_t *gl_cull;
cvar_t *gl_playermip;
cvar_t *gl_saturatelighting;
cvar_t *gl_texturemode;
cvar_t *gl_texturealphamode;
cvar_t *gl_texturesolidmode;
cvar_t *gl_lockpvs;

cvar_t *gl_3dlabs_broken;

cvar_t *vid_width;
cvar_t *vid_height;
cvar_t *vid_fullscreen;
cvar_t *vid_gamma;
cvar_t *vid_ref;

cvar_t *skydistance;

/*
R_CullBox
 
Returns true if the box is completely outside the frustom
*/
qboolean R_CullBox(vec3_t mins, vec3_t maxs){
	int i;
	
	if(r_nocull->value)
		return false;
		
	for(i = 0; i < 4; i++)
		if(BOX_ON_PLANE_SIDE(mins, maxs, &frustum[i]) == 2)
			return true;
	return false;
}


void R_RotateForEntity(entity_t *e){
	qglTranslatef(e->origin[0], e->origin[1], e->origin[2]);
	
	qglRotatef(e->angles[1], 0, 0, 1);
	qglRotatef(-e->angles[0], 0, 1, 0);
	qglRotatef(-e->angles[2], 1, 0, 0);
}

/*
 
  SPRITE MODELS
 
*/


/*
R_DrawSpriteModel
 
*/
void R_DrawSpriteModel(entity_t *e){
	float alpha = 1.0F;
	vec3_t point;
	dsprframe_t *frame;
	float *up, *right;
	dsprite_t *psprite;
	
	// don't even bother culling, because it's just a single
	// polygon without a surface cache
	
	psprite =(dsprite_t *)currentmodel->extradata;
	
	e->frame %= psprite->numframes;
	
	frame = &psprite->frames[e->frame];
	
	up = vup;
	right = vright;
	
	if(e->flags & RF_TRANSLUCENT)
		alpha = e->alpha;
		
	if(alpha != 1.0F){
		qglEnable(GL_BLEND);
		color_white[3] = alpha * 255;
	}
	qglColor4ubv(color_white);
	
	GL_Bind(currentmodel->skins[e->frame]->texnum);
	
	GL_TexEnv(GL_MODULATE);
	
	if(alpha == 1.0)
		qglEnable(GL_ALPHA_TEST);
	else
		qglDisable(GL_ALPHA_TEST);
		
	qglBegin(GL_QUADS);
	
	qglTexCoord2f(0, 1);
	VectorMA(e->origin, -frame->origin_y, up, point);
	VectorMA(point, -frame->origin_x, right, point);
	qglVertex3fv(point);
	
	qglTexCoord2f(0, 0);
	VectorMA(e->origin, frame->height - frame->origin_y, up, point);
	VectorMA(point, -frame->origin_x, right, point);
	qglVertex3fv(point);
	
	qglTexCoord2f(1, 0);
	VectorMA(e->origin, frame->height - frame->origin_y, up, point);
	VectorMA(point, frame->width - frame->origin_x, right, point);
	qglVertex3fv(point);
	
	qglTexCoord2f(1, 1);
	VectorMA(e->origin, -frame->origin_y, up, point);
	VectorMA(point, frame->width - frame->origin_x, right, point);
	qglVertex3fv(point);
	
	qglEnd();
	
	qglDisable(GL_ALPHA_TEST);
	GL_TexEnv(GL_REPLACE);
	
	if(alpha != 1.0F){
		qglDisable(GL_BLEND);
		color_white[3] = 255;
	}
	qglColor4ubv(color_white);
}


/*
R_DrawNullModel
*/
void R_DrawNullModel(void){
	vec3_t shadelight;
	int i;
	
	if(currententity->flags & RF_FULLBRIGHT)
		shadelight[0] = shadelight[1] = shadelight[2] = 1.0F;
	else
		R_LightPoint(currententity->origin, shadelight);
		
	qglPushMatrix();
	R_RotateForEntity(currententity);
	
	qglDisable(GL_TEXTURE_2D);
	qglColor3fv(shadelight);
	
	qglBegin(GL_TRIANGLE_FAN);
	qglVertex3f(0, 0, -16);
	for(i = 0; i <= 4; i++)
		qglVertex3f(16*cos(i*M_PI / 2), 16*sin(i*M_PI / 2), 0);
	qglEnd();
	
	qglBegin(GL_TRIANGLE_FAN);
	qglVertex3f(0, 0, 16);
	for(i = 4; i >= 0; i--)
		qglVertex3f(16*cos(i*M_PI / 2), 16*sin(i*M_PI / 2), 0);
	qglEnd();
	
	qglColor3ubv(color_white);
	qglPopMatrix();
	qglEnable(GL_TEXTURE_2D);
}

/*
R_DrawEntitiesOnList
*/
void R_DrawEntitiesOnList(void){
	int i;
	
	if(!r_drawentities->value)
		return;
		
	// draw non-transparent first
	for(i = 0; i < r_newrefdef.num_entities; i++){
		currententity = &r_newrefdef.entities[i];
		if(currententity->flags & RF_TRANSLUCENT)
			continue;  // solid
			
		if(currententity->flags & RF_BEAM){
			R_DrawBeam(currententity);
		} else {
			currentmodel = currententity->model;
			if(!currentmodel){
				R_DrawNullModel();
				continue;
			}
			switch(currentmodel->type){
				case mod_alias:
					R_DrawAliasModel(currententity);
					break;
				case mod_brush:
					R_DrawBrushModel(currententity);
					break;
				case mod_sprite:
					R_DrawSpriteModel(currententity);
					break;
				default:
					ri.Sys_Error(ERR_DROP, "Bad modeltype");
					break;
			}
		}
	}
	
	// draw transparent entities
	// we could sort these if it ever becomes a problem...
	qglDepthMask(0);  // no z writes
	for(i = 0; i < r_newrefdef.num_entities; i++){
		currententity = &r_newrefdef.entities[i];
		if(!(currententity->flags & RF_TRANSLUCENT))
			continue;  // solid
			
		if(currententity->flags & RF_BEAM){
			R_DrawBeam(currententity);
		} else {
			currentmodel = currententity->model;
			
			if(!currentmodel){
				R_DrawNullModel();
				continue;
			}
			switch(currentmodel->type){
				case mod_alias:
					R_DrawAliasModel(currententity);
					break;
				case mod_brush:
					R_DrawBrushModel(currententity);
					break;
				case mod_sprite:
					R_DrawSpriteModel(currententity);
					break;
				default:
					ri.Sys_Error(ERR_DROP, "Bad modeltype");
					break;
			}
		}
	}
	qglDepthMask(1);  // back to writing
	
}

/*
R_DrawParticles
*/
void R_DrawParticles(void){
	const particle_t *p;
	int i;
	vec3_t up, right, up_right_scale, down_right_scale, VA[4];
	float scale;
	byte color[4];
	
	GL_Bind(r_particletexture->texnum);
	//qglDepthMask(GL_FALSE);  // no z buffering
	qglEnable(GL_BLEND);
	GL_TexEnv(GL_MODULATE);
	qglBegin(GL_QUADS);
	
	for(p = r_newrefdef.particles, i = 0; i < r_newrefdef.num_particles; i++, p++){
	// hack a scale up to keep particles from disapearing
		scale =(p->origin[0] - r_origin[0]) * vpn[0] +
				(p->origin[1] - r_origin[1]) * vpn[1] +
				(p->origin[2] - r_origin[2]) * vpn[2];
		        
		scale *= 0.002;
		
		VectorScale(vup, scale, up);
		VectorScale(vright, scale, right);
		VectorAdd(up, right, up_right_scale);
		VectorSubtract(right, up, down_right_scale);
		VectorSubtract(p->origin, down_right_scale, VA[0]);
		VectorAdd(p->origin, up_right_scale, VA[1]);
		VectorAdd(p->origin, down_right_scale, VA[2]);
		VectorSubtract(p->origin, up_right_scale, VA[3]);
		
		*(int *)color = d_8to24table[p->color];
		color[3] = p->alpha * 255;
		qglColor4ubv(color);
		
		qglTexCoord2f(0.0, 0.0);
		qglVertex3fv(VA[0]);
		
		qglTexCoord2f(1.0, 0.0);
		qglVertex3fv(VA[1]);
		
		qglTexCoord2f(1.0, 1.0);
		qglVertex3fv(VA[2]);
		
		qglTexCoord2f(0.0, 1.0);
		qglVertex3fv(VA[3]);
	}
	
	qglEnd();
	qglDisable(GL_BLEND);
	qglColor4ubv(color_white);
	//qglDepthMask(GL_TRUE);  // back to normal Z buffering
	GL_TexEnv(GL_REPLACE);
}

int SignbitsForPlane(cplane_t *out){
	int bits, j;
	
	// for fast box on planeside test
	
	bits = 0;
	for(j = 0; j < 3; j++){
		if(out->normal[j] < 0)
			bits |= 1 << j;
	}
	return bits;
}


void R_SetFrustum(void){
	int i;
	
	// rotate VPN right by FOV_X/2 degrees
	RotatePointAroundVector(frustum[0].normal, vup, vpn, -(90 - r_newrefdef.fov_x / 2));
	// rotate VPN left by FOV_X/2 degrees
	RotatePointAroundVector(frustum[1].normal, vup, vpn, 90 - r_newrefdef.fov_x / 2);
	// rotate VPN up by FOV_X/2 degrees
	RotatePointAroundVector(frustum[2].normal, vright, vpn, 90 - r_newrefdef.fov_y / 2);
	// rotate VPN down by FOV_X/2 degrees
	RotatePointAroundVector(frustum[3].normal, vright, vpn, -(90 - r_newrefdef.fov_y / 2));
	
	for(i = 0; i < 4; i++){
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct(r_origin, frustum[i].normal);
		frustum[i].signbits = SignbitsForPlane(&frustum[i]);
	}
}


/*
R_SetupFrame
*/
void R_SetupFrame(void){
	int i;
	mleaf_t *leaf;
	
	r_framecount++;
	
	// build the transformation matrix for the given view angles
	VectorCopy(r_newrefdef.vieworg, r_origin);
	
	AngleVectors(r_newrefdef.viewangles, vpn, vright, vup);
	
	// current viewcluster
	if(!(r_newrefdef.rdflags & RDF_NOWORLDMODEL)){
		r_oldviewcluster = r_viewcluster;
		r_oldviewcluster2 = r_viewcluster2;
		leaf = Mod_PointInLeaf(r_origin, r_worldmodel);
		r_viewcluster = r_viewcluster2 = leaf->cluster;
		
		// check above and below so crossing solid water doesn't draw wrong
		if(!leaf->contents){  // look down a bit
			vec3_t temp;
			
			VectorCopy(r_origin, temp);
			temp[2] -= 16;
			leaf = Mod_PointInLeaf(temp, r_worldmodel);
			if(!(leaf->contents & CONTENTS_SOLID) &&
					(leaf->cluster != r_viewcluster2))
				r_viewcluster2 = leaf->cluster;
		} else {  // look up a bit
			vec3_t temp;
			
			VectorCopy(r_origin, temp);
			temp[2] += 16;
			leaf = Mod_PointInLeaf(temp, r_worldmodel);
			if(!(leaf->contents & CONTENTS_SOLID) &&
					(leaf->cluster != r_viewcluster2))
				r_viewcluster2 = leaf->cluster;
		}
	}
	
	for(i = 0; i < 4; i++)
		v_blend[i] = r_newrefdef.blend[i];
		
	c_brush_polys = 0;
	c_alias_polys = 0;
}


void MYgluPerspective(GLdouble fovy, GLdouble aspect,
					   GLdouble zNear, GLdouble zFar){
	GLdouble xmin, xmax, ymin, ymax;
	
	ymax = zNear * tan(fovy * M_PI / 360.0);
	ymin = -ymax;
	
	xmin = ymin * aspect;
	xmax = ymax * aspect;
	
	qglFrustum(xmin, xmax, ymin, ymax, zNear, zFar);
}


/*
R_SetupGL
*/
void R_SetupGL(void){
	float screenaspect;
	//	float yfov;
	int x, x2, y2, y, w, h;
	/* skybox vars */
	static GLdouble farz;
	GLdouble boxsize;
	
	//
	// set up viewport
	//
	x = floor(r_newrefdef.x * viddef.width / viddef.width);
	x2 = ceil((r_newrefdef.x + r_newrefdef.width) * viddef.width / viddef.width);
	y = floor(viddef.height - r_newrefdef.y * viddef.height / viddef.height);
	y2 = ceil(viddef.height -(r_newrefdef.y + r_newrefdef.height) * viddef.height / viddef.height);
	
	w = x2 - x;
	h = y - y2;
	
	qglViewport(x, y2, w, h);
	
	/* block of code for SkyBox Size change */
	if(skydistance->modified){
		skydistance->modified = false;
		boxsize = skydistance->value;
		boxsize -= 252 * ceil(boxsize / 2300);
		farz = 1.0;
		while(farz < boxsize){ /* make this value a power-of-2 */
			farz *= 2.0;
			if(farz >= 65536.0)  /* don't make it larger than this */
				break;
		}
		/* double since boxsize is distance from camera to
		 * edge of skybox - not total size of skybox */
		farz *= 2.0;
		ri.Con_Printf(PRINT_DEVELOPER, "farz now set to %g\n", farz);
	}
	
	//
	// set up projection matrix
	//
	screenaspect =(float)r_newrefdef.width / r_newrefdef.height;
	qglMatrixMode(GL_PROJECTION);
	qglLoadIdentity();
	MYgluPerspective(r_newrefdef.fov_y, screenaspect, 4, farz);
	
	qglCullFace(GL_FRONT);
	
	qglMatrixMode(GL_MODELVIEW);
	qglLoadIdentity();
	
	qglRotatef(-90, 1, 0, 0);	    // put Z going up
	qglRotatef(90, 0, 0, 1);	    // put Z going up
	qglRotatef(-r_newrefdef.viewangles[2], 1, 0, 0);
	qglRotatef(-r_newrefdef.viewangles[0], 0, 1, 0);
	qglRotatef(-r_newrefdef.viewangles[1], 0, 0, 1);
	qglTranslatef(-r_newrefdef.vieworg[0], -r_newrefdef.vieworg[1], -r_newrefdef.vieworg[2]);
	
	qglGetFloatv(GL_MODELVIEW_MATRIX, r_world_matrix);
	
	//
	// set drawing parms
	//
	if(gl_cull->value)
		qglEnable(GL_CULL_FACE);
	else
		qglDisable(GL_CULL_FACE);
		
	qglDisable(GL_BLEND);
	qglDisable(GL_ALPHA_TEST);
	qglEnable(GL_DEPTH_TEST);
}


/*
R_Clear
*/
void R_Clear(void){
	if(gl_ztrick->value && r_worldmodel){
		static int trickframe;
		
		if(gl_clear->value)
			qglClear(GL_COLOR_BUFFER_BIT);
			
		trickframe++;
		if(trickframe & 1){
			gldepthmin = 0;
			gldepthmax = 0.49999;
			qglDepthFunc(GL_LEQUAL);
		} else {
			gldepthmin = 1;
			gldepthmax = 0.5;
			qglDepthFunc(GL_GEQUAL);
		}
	} else {
		if(gl_clear->value)
			qglClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		else
			qglClear(GL_DEPTH_BUFFER_BIT);
		gldepthmin = 0;
		gldepthmax = 1;
		qglDepthFunc(GL_LEQUAL);
	}
	
	qglDepthRange(gldepthmin, gldepthmax);
}

/*
R_RenderView
 
r_newrefdef must be set before the first call
*/
void R_RenderView(refdef_t *fd){
	if(r_norefresh->value)
		return;
		
	r_newrefdef = *fd;
	
	if(!r_worldmodel && !(r_newrefdef.rdflags & RDF_NOWORLDMODEL))
		ri.Sys_Error(ERR_DROP, "R_RenderView: NULL worldmodel");
		
	if(r_speeds->value){
		c_brush_polys = 0;
		c_alias_polys = 0;
	}
	
	if(gl_finish->value)
		qglFinish();
		
	R_SetupFrame();
	
	R_SetFrustum();
	
	R_SetupGL();
	
	R_MarkLeaves();  // done here so we know if we're in water
	
	if(gl_wireframe->value){
		qglPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		qglClear(GL_COLOR_BUFFER_BIT);
	}
	
	R_DrawWorld();
	
	if(gl_wireframe->value)
		qglPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	
	R_DrawEntitiesOnList();
	
	R_DrawParticles();
	
	R_DrawAlphaSurfaces();
	
	if(r_speeds->value){
		ri.Con_Printf(PRINT_ALL, "%4i wpoly %4i epoly %i tex %i lmaps\n",
					   c_brush_polys,
					   c_alias_polys,
					   c_visible_textures,
					   c_visible_lightmaps);
	}
}

void	R_SetGL2D(void){
	// set 2D virtual screen size
	qglViewport(0, 0, viddef.width, viddef.height);
	qglMatrixMode(GL_PROJECTION);
	qglLoadIdentity();
	qglOrtho(0, viddef.width, viddef.height, 0, -99999, 99999);
	qglMatrixMode(GL_MODELVIEW);
	qglLoadIdentity();
	qglDisable(GL_DEPTH_TEST);
	qglDisable(GL_CULL_FACE);
	qglDisable(GL_BLEND);
	qglEnable(GL_ALPHA_TEST);
	qglColor4f(1, 1, 1, 0.3);
}

/*
R_SetLightLevel
 
*/
void R_SetLightLevel(void){
	vec3_t shadelight;
	
	if(r_newrefdef.rdflags & RDF_NOWORLDMODEL)
		return;
		
	// save off light value for server to look at(BIG HACK!)
	
	R_LightPoint(r_newrefdef.vieworg, shadelight);
	
	// pick the greatest component, which should be the same
	// as the mono value returned by software
	if(shadelight[0] > shadelight[1]){
		if(shadelight[0] > shadelight[2])
			r_lightlevel->value = 150 * shadelight[0];
		else
			r_lightlevel->value = 150 * shadelight[2];
	} else {
		if(shadelight[1] > shadelight[2])
			r_lightlevel->value = 150 * shadelight[1];
		else
			r_lightlevel->value = 150 * shadelight[2];
	}
	
}

/*
@@@@@@@@@@@@@@@@@@@@@
R_RenderFrame
 
@@@@@@@@@@@@@@@@@@@@@
*/
void R_RenderFrame(refdef_t *fd){
	R_RenderView(fd);
	R_SetLightLevel();
	R_SetGL2D();
}


void R_Register(void){
	r_norefresh = ri.Cvar_Get("r_norefresh", "0", 0);
	r_fullbright = ri.Cvar_Get("r_fullbright", "0", 0);
	r_drawentities = ri.Cvar_Get("r_drawentities", "1", 0);
	r_drawworld = ri.Cvar_Get("r_drawworld", "1", 0);
	r_novis = ri.Cvar_Get("r_novis", "0", 0);
	r_nocull = ri.Cvar_Get("r_nocull", "0", 0);
	r_lerpmodels = ri.Cvar_Get("r_lerpmodels", "1", 0);
	r_speeds = ri.Cvar_Get("r_speeds", "0", 0);
	
	r_lightlevel = ri.Cvar_Get("r_lightlevel", "0", 0);
	
	gl_nosubimage = ri.Cvar_Get("gl_nosubimage", "0", 0);
	gl_allow_software = ri.Cvar_Get("gl_allow_software", "0", 0);
	
	gl_modulate = ri.Cvar_Get("gl_modulate", "1", CVAR_ARCHIVE);
	gl_log = ri.Cvar_Get("gl_log", "0", 0);
	gl_bitdepth = ri.Cvar_Get("gl_bitdepth", "0", 0);
	gl_lightmap = ri.Cvar_Get("gl_lightmap", "0", 0);
	gl_nobind = ri.Cvar_Get("gl_nobind", "0", 0);
	gl_round_down = ri.Cvar_Get("gl_round_down", "1", 0);
	gl_picmip = ri.Cvar_Get("gl_picmip", "0", 0);
	gl_skymip = ri.Cvar_Get("gl_skymip", "0", 0);
	gl_wireframe = ri.Cvar_Get("gl_wireframe", "0", 0);
	gl_ztrick = ri.Cvar_Get("gl_ztrick", "1", CVAR_ARCHIVE);
	gl_finish = ri.Cvar_Get("gl_finish", "0", CVAR_ARCHIVE);
	gl_clear = ri.Cvar_Get("gl_clear", "0", 0);
	gl_cull = ri.Cvar_Get("gl_cull", "1", 0);
	gl_playermip = ri.Cvar_Get("gl_playermip", "0", 0);
	
	gl_driver = ri.Cvar_Get("gl_driver", "libGL.so.1", CVAR_ARCHIVE);
	
	gl_texturemode = ri.Cvar_Get("gl_texturemode", "GL_LINEAR_MIPMAP_NEAREST", CVAR_ARCHIVE);
	gl_texturealphamode = ri.Cvar_Get("gl_texturealphamode", "default", CVAR_ARCHIVE);
	gl_texturesolidmode = ri.Cvar_Get("gl_texturesolidmode", "default", CVAR_ARCHIVE);
	gl_lockpvs = ri.Cvar_Get("gl_lockpvs", "0", 0);
	
	gl_vertex_arrays = ri.Cvar_Get("gl_vertex_arrays", "1", CVAR_ARCHIVE);
	
	gl_ext_palettedtexture = ri.Cvar_Get("gl_ext_palettedtexture", "1", CVAR_ARCHIVE);
	gl_ext_multitexture = ri.Cvar_Get("gl_ext_multitexture", "1", CVAR_ARCHIVE);
	gl_ext_compiled_vertex_array = ri.Cvar_Get("gl_ext_compiled_vertex_array", "1", CVAR_ARCHIVE);
	
	gl_drawbuffer = ri.Cvar_Get("gl_drawbuffer", "GL_BACK", 0);
	
	gl_saturatelighting = ri.Cvar_Get("gl_saturatelighting", "0", 0);
	
	gl_3dlabs_broken = ri.Cvar_Get("gl_3dlabs_broken", "1", CVAR_ARCHIVE);
	
	vid_width = ri.Cvar_Get("vid_width", "640", CVAR_ARCHIVE);
	vid_height = ri.Cvar_Get("vid_height", "480", CVAR_ARCHIVE);
	vid_fullscreen = ri.Cvar_Get("vid_fullscreen", "0", CVAR_ARCHIVE);
	vid_gamma = ri.Cvar_Get("vid_gamma", "1.0", CVAR_ARCHIVE);
	vid_ref = ri.Cvar_Get("vid_ref", "glx", CVAR_ARCHIVE);
	
	skydistance = ri.Cvar_Get("skydistance", "2300", CVAR_ARCHIVE);
	
	ri.Cmd_AddCommand("imagelist", GL_ImageList_f);
	ri.Cmd_AddCommand("screenshot", GL_ScreenShot_f);
	ri.Cmd_AddCommand("modellist", Mod_Modellist_f);
	ri.Cmd_AddCommand("gl_strings", GL_Strings_f);
}

/*
R_SetMode
*/
qboolean R_SetMode(void){
	
	vid_width->modified = false;
	vid_height->modified = false;
	vid_fullscreen->modified = false;
	
	if(GLimp_SetMode(vid_width->value, vid_height->value, vid_fullscreen->value)){
		
		viddef.width = (int)vid_width->value;
		viddef.height = (int)vid_height->value;
		
		gl_state.prev_width = vid_width->value;
		gl_state.prev_height = vid_height->value;
		
		skydistance->modified = true;
		return true;  //succeded
	}
	
	ri.Cvar_SetValue("vid_width", gl_state.prev_width);  //failed, revert
	ri.Cvar_SetValue("vid_height", gl_state.prev_height);
	
	GLimp_SetMode(gl_state.prev_width, gl_state.prev_height, false);
	return false;
}

/*
R_Init
*/
qboolean R_Init(void *hinstance, void *hWnd){
	char renderer_buffer[MAX_STRING_CHARS];
	char vendor_buffer[MAX_STRING_CHARS];
	int j, err;
	
	extern float r_turbsin[256];
	for(j = 0; j < 256; j++){
		r_turbsin[j] *= 0.5;
	}
	
	ri.Con_Printf(PRINT_ALL, "ref_gl version: "REF_VERSION"\n");
	
	Draw_GetPalette();
	
	R_Register();
	
	// initialize our QGL dynamic bindings
	if(!QGL_Init(gl_driver->string)){
		ri.Con_Printf(PRINT_ALL, "GL: Failed to load driver(%s)\n", gl_driver->string);
		QGL_Shutdown();
		return -1;
	}
	
	// initialize OS-specific parts of OpenGL
	if(!GLimp_Init(hinstance, hWnd)){
		ri.Con_Printf(PRINT_ALL, "GL: Failed to initialize context\n");
		QGL_Shutdown();
		return -1;
	}
	
	gl_state.prev_width = 640;
	gl_state.prev_height = 480;
	
	// create the window and set up the context
	if(!R_SetMode()){
		ri.Con_Printf(PRINT_ALL, "GL: Failed to set video mode\n");
		QGL_Shutdown();
		return -1;
	}
	
	gl_config.vendor_string =(char *)qglGetString(GL_VENDOR);
	gl_config.renderer_string =(char *)qglGetString(GL_RENDERER);
	gl_config.version_string =(char *)qglGetString(GL_VERSION);
	gl_config.extensions_string =(char *)qglGetString(GL_EXTENSIONS);
	
	strncpy(renderer_buffer, gl_config.renderer_string, sizeof(renderer_buffer));
	strncpy(vendor_buffer, gl_config.vendor_string, sizeof(vendor_buffer));
	
	strlwr(renderer_buffer); strlwr(vendor_buffer);
	
	if(strstr(renderer_buffer, "voodoo")){  //resolve renderer
		if(!strstr(renderer_buffer, "rush"))
			gl_config.renderer = GL_RENDERER_VOODOO;
		else
			gl_config.renderer = GL_RENDERER_VOODOO_RUSH;
	} else if(strstr(vendor_buffer, "sgi"))
		gl_config.renderer = GL_RENDERER_SGI;
	else if(strstr(renderer_buffer, "permedia"))
		gl_config.renderer = GL_RENDERER_PERMEDIA2;
	else if(strstr(renderer_buffer, "glint"))
		gl_config.renderer = GL_RENDERER_GLINT_MX;
	else if(strstr(renderer_buffer, "glzicd"))
		gl_config.renderer = GL_RENDERER_REALIZM;
	else if(strstr(renderer_buffer, "gdi"))
		gl_config.renderer = GL_RENDERER_MCD;
	else if(strstr(renderer_buffer, "pcx2"))
		gl_config.renderer = GL_RENDERER_PCX2;
	else if(strstr(renderer_buffer, "verite"))
		gl_config.renderer = GL_RENDERER_RENDITION;
	else gl_config.renderer = GL_RENDERER_OTHER;
	
	// MCD has buffering issues
	if(gl_config.renderer == GL_RENDERER_MCD){
		ri.Cvar_SetValue("gl_finish", 1);
	}
	
	if(gl_config.renderer & GL_RENDERER_3DLABS){
		if(gl_3dlabs_broken->value) gl_config.allow_cds = false;
		else gl_config.allow_cds = true;
	}
	else gl_config.allow_cds = true;
	
	// enable available extensions
	if(strstr(gl_config.extensions_string, "GL_EXT_compiled_vertex_array") ||
			strstr(gl_config.extensions_string, "GL_SGI_compiled_vertex_array")){
		qglLockArraysEXT = (void(APIENTRY*)(int, int)) qwglGetProcAddress("glLockArraysEXT");
		qglUnlockArraysEXT = (void(APIENTRY*)(void)) qwglGetProcAddress("glUnlockArraysEXT");
	//Com_Printf("GL_EXT_compiled_vertex_array\n");
	}
	
	if(strstr(gl_config.extensions_string, "3DFX_set_global_palette") && gl_ext_palettedtexture->value){
		qgl3DfxSetPaletteEXT = (void(APIENTRY *)(GLuint *))qwglGetProcAddress("gl3DfxSetPaletteEXT");
		qglColorTableEXT = Fake_glColorTableEXT;
	//Com_Printf("3DFX_set_global_palette\n");
	}
	
	if(!qglColorTableEXT && strstr(gl_config.extensions_string, "GL_EXT_paletted_texture") &&
			strstr(gl_config.extensions_string, "GL_EXT_shared_texture_palette") && 
			gl_ext_palettedtexture->value){
		qglColorTableEXT = (void(APIENTRY *)(GLenum, GLenum, GLsizei, GLenum, GLenum, const GLvoid*))
				qwglGetProcAddress("glColorTableEXT");
	//Com_Printf("GL_EXT_shared_texture_palette\n");
	}
	
	if(strstr(gl_config.extensions_string, "GL_ARB_multitexture") && gl_ext_multitexture->value){
		qglMTexCoord2fSGIS = (void(APIENTRY*)(GLenum, GLfloat, GLfloat)) qwglGetProcAddress("glMultiTexCoord2fARB");
		qglActiveTextureARB = (void(APIENTRY*)(GLenum)) qwglGetProcAddress("glActiveTextureARB");
		qglClientActiveTextureARB =(void(APIENTRY*)(GLenum)) qwglGetProcAddress("glClientActiveTextureARB");
		QGL_TEXTURE0 = GL_TEXTURE0_ARB;
		QGL_TEXTURE1 = GL_TEXTURE1_ARB;
	//Com_Printf("GL_ARB_multitexture\n");
	}
	
	if(strstr(gl_config.extensions_string, "GL_SGIS_multitexture")){
		if(!qglActiveTextureARB && gl_ext_multitexture->value){
			qglMTexCoord2fSGIS = (void(APIENTRY*)(GLenum, GLfloat, GLfloat)) qwglGetProcAddress("glMTexCoord2fSGIS");
			qglSelectTextureSGIS = (void(APIENTRY*)(GLenum)) qwglGetProcAddress("glSelectTextureSGIS");
			QGL_TEXTURE0 = GL_TEXTURE0_SGIS;
			QGL_TEXTURE1 = GL_TEXTURE1_SGIS;
	//Com_Printf("GL_SGIS_multitexture\n");
		}
	}
	
	GL_SetDefaultState();
	GL_InitImages();
	
	Mod_Init();
	R_InitParticleTexture();
	R_InitNoTexture();
	Draw_InitLocal();
	
	err = qglGetError();
	if(err != GL_NO_ERROR)
		ri.Con_Printf(PRINT_ALL, "glGetError() = 0x%x\n", err);
	return 0;
}

/*
R_Shutdown
*/
void R_Shutdown(void){
	ri.Cmd_RemoveCommand("modellist");
	ri.Cmd_RemoveCommand("screenshot");
	ri.Cmd_RemoveCommand("imagelist");
	ri.Cmd_RemoveCommand("gl_strings");
	
	Mod_FreeAll();
	
	GL_ShutdownImages();
	
	/*
	** shut down OS specific OpenGL stuff like contexts, etc.
	*/
	GLimp_Shutdown();
	
	/*
	** shutdown our QGL subsystem
	*/
	QGL_Shutdown();
}


/*
*  Implemented per ref?
*/
extern void UpdateHardwareGamma();


/*
@@@@@@@@@@@@@@@@@@@@@
R_BeginFrame
@@@@@@@@@@@@@@@@@@@@@
*/
void R_BeginFrame(void){

	if(gl_log->modified){
		GLimp_EnableLogging(gl_log->value);
		gl_log->modified = false;
	}
	
	if(gl_log->value){
		GLimp_LogNewFrame();
	}
	
	/*
	** update 3Dfx gamma -- it is expected that a user will do a vid_restart
	** after tweaking this value
	*/
	if(vid_gamma->modified){
		vid_gamma->modified = false;
		
		if(gl_state.hwgamma){
			UpdateHardwareGamma();
		} else if(gl_config.renderer & (GL_RENDERER_VOODOO)){
			char envbuffer[1024];
			float g;
			
			g = 2.00 *(0.8 -(vid_gamma->value - 0.5)) + 1.0F;
			Com_sprintf(envbuffer, sizeof(envbuffer), "SSTV2_GAMMA=%f", g);
			putenv(envbuffer);
			Com_sprintf(envbuffer, sizeof(envbuffer), "SST_GAMMA=%f", g);
			putenv(envbuffer);
		}
	}
	
	GLimp_BeginFrame();
	
	/*
	** go into 2D mode
	*/
	qglViewport(0, 0, viddef.width, viddef.height);
	qglMatrixMode(GL_PROJECTION);
	qglLoadIdentity();
	qglOrtho(0, viddef.width, viddef.height, 0, -99999, 99999);
	qglMatrixMode(GL_MODELVIEW);
	qglLoadIdentity();
	qglDisable(GL_DEPTH_TEST);
	qglDisable(GL_CULL_FACE);
	qglDisable(GL_BLEND);
	qglEnable(GL_ALPHA_TEST);
	qglEnable(GL_TEXTURE_2D);
	qglColor4ubv(color_white);
	
	/*
	** draw buffer stuff
	*/
	if(gl_drawbuffer->modified){
		gl_drawbuffer->modified = false;
		
		if(!Q_stricmp(gl_drawbuffer->string, "GL_FRONT"))
			qglDrawBuffer(GL_FRONT);
		else qglDrawBuffer(GL_BACK);
	}
	
	/*
	** texturemode stuff
	*/
	if(gl_texturemode->modified){
		GL_TextureMode(gl_texturemode->string);
		gl_texturemode->modified = false;
	}
	
	if(gl_texturealphamode->modified){
		GL_TextureAlphaMode(gl_texturealphamode->string);
		gl_texturealphamode->modified = false;
	}
	
	if(gl_texturesolidmode->modified){
		GL_TextureSolidMode(gl_texturesolidmode->string);
		gl_texturesolidmode->modified = false;
	}
	
	// clear screen if desired
	R_Clear();
}

/*
R_SetPalette
*/
unsigned r_rawpalette[256];

void R_SetPalette(const unsigned char *palette){
	int i;
	
	byte *rp =(byte *) r_rawpalette;
	
	if(palette){
		for(i = 0; i < 256; i++){
			rp[i*4 + 0] = palette[i * 3 + 0];
			rp[i*4 + 1] = palette[i * 3 + 1];
			rp[i*4 + 2] = palette[i * 3 + 2];
			rp[i*4 + 3] = 0xff;
		}
	} else {
		for(i = 0; i < 256; i++){
			rp[i*4 + 0] = d_8to24table[i] & 0xff;
			rp[i*4 + 1] = (d_8to24table[i] >> 8) & 0xff;
			rp[i*4 + 2] = (d_8to24table[i] >> 16) & 0xff;
			rp[i*4 + 3] = 0xff;
		}
	}
	GL_SetTexturePalette(r_rawpalette);
	
	qglClear(GL_COLOR_BUFFER_BIT);
}

/*
** R_DrawBeam
*/
void R_DrawBeam(entity_t *e){
#define NUM_BEAM_SEGS 6

	int i;
	float r, g, b;
	
	vec3_t perpvec;
	vec3_t direction, normalized_direction;
	vec3_t start_points[NUM_BEAM_SEGS], end_points[NUM_BEAM_SEGS];
	vec3_t oldorigin, origin;
	
	oldorigin[0] = e->oldorigin[0];
	oldorigin[1] = e->oldorigin[1];
	oldorigin[2] = e->oldorigin[2];
	
	origin[0] = e->origin[0];
	origin[1] = e->origin[1];
	origin[2] = e->origin[2];
	
	normalized_direction[0] = direction[0] = oldorigin[0] - origin[0];
	normalized_direction[1] = direction[1] = oldorigin[1] - origin[1];
	normalized_direction[2] = direction[2] = oldorigin[2] - origin[2];
	
	if(VectorNormalize(normalized_direction) == 0)
		return;
		
	PerpendicularVector(perpvec, normalized_direction);
	VectorScale(perpvec, e->frame / 2, perpvec);
	
	for(i = 0; i < 6; i++){
		RotatePointAroundVector(start_points[i], normalized_direction, perpvec,(360.0 / NUM_BEAM_SEGS)*i);
		VectorAdd(start_points[i], origin, start_points[i]);
		VectorAdd(start_points[i], direction, end_points[i]);
	}
	
	qglDisable(GL_TEXTURE_2D);
	qglEnable(GL_BLEND);
	qglDepthMask(GL_FALSE);
	
	r =(d_8to24table[e->skinnum & 0xFF]) & 0xFF;
	g =(d_8to24table[e->skinnum & 0xFF] >> 8) & 0xFF;
	b =(d_8to24table[e->skinnum & 0xFF] >> 16) & 0xFF;
	
	r *= 1 / 255.0F;
	g *= 1 / 255.0F;
	b *= 1 / 255.0F;
	
	qglColor4f(r, g, b, e->alpha);
	
	qglBegin(GL_TRIANGLE_STRIP);
	for(i = 0; i < NUM_BEAM_SEGS; i++){
		qglVertex3fv(start_points[i]);
		qglVertex3fv(end_points[i]);
		qglVertex3fv(start_points[(i + 1) % NUM_BEAM_SEGS]);
		qglVertex3fv(end_points[(i + 1) % NUM_BEAM_SEGS]);
	}
	qglEnd();
	
	qglEnable(GL_TEXTURE_2D);
	qglDisable(GL_BLEND);
	qglDepthMask(GL_TRUE);
}



void	R_BeginRegistration(char *map);
struct model_s	*R_RegisterModel(char *name);
struct image_s	*R_RegisterSkin(char *name);
void R_SetSky(char *name, float rotate, vec3_t axis);
void	R_EndRegistration(void);

void	R_RenderFrame(refdef_t *fd);

struct image_s	*Draw_FindPic(char *name);

void	Draw_Pic(int x, int y, char *name);
void	Draw_Char(int x, int y, int c);
void	Draw_TileClear(int x, int y, int w, int h, char *name);
void	Draw_Fill(int x, int y, int w, int h, int c);
void	Draw_FadeScreen(void);

/*
@@@@@@@@@@@@@@@@@@@@@
GetRefAPI
 
@@@@@@@@@@@@@@@@@@@@@
*/
refexport_t GetRefAPI(refimport_t rimp){
	refexport_t re;
	
	ri = rimp;
	
	re.api_version = API_VERSION;
	
	re.BeginRegistration = R_BeginRegistration;
	re.RegisterModel = R_RegisterModel;
	re.RegisterSkin = R_RegisterSkin;
	re.RegisterPic = Draw_FindPic;
	re.SetSky = R_SetSky;
	re.EndRegistration = R_EndRegistration;
	
	re.RenderFrame = R_RenderFrame;
	
	re.DrawGetPicSize = Draw_GetPicSize;
	re.DrawPic = Draw_Pic;
	re.DrawStretchPic = Draw_StretchPic;
	re.DrawChar = Draw_Char;
	re.DrawTileClear = Draw_TileClear;
	re.DrawFill = Draw_Fill;
	re.DrawFadeScreen = Draw_FadeScreen;
	
	re.DrawStretchRaw = Draw_StretchRaw;
	
	re.Init = R_Init;
	re.Shutdown = R_Shutdown;
	
	re.CinematicSetPalette = R_SetPalette;
	re.BeginFrame = R_BeginFrame;
	re.EndFrame = GLimp_EndFrame;
	
	re.AppActivate = GLimp_AppActivate;
	
	Swap_Init();
	
	return re;
}


#ifndef REF_HARD_LINKED 
// this is only here so the functions in q_shared.c and q_sh.c can link
void Sys_Error(char *error, ...){
	va_list	argptr;
	char text[1024];
	
	va_start(argptr, error);
	vsprintf(text, error, argptr);
	va_end(argptr);
	
	ri.Sys_Error(ERR_FATAL, "%s", text);
}

void Com_Printf(char *fmt, ...){
	va_list	argptr;
	char text[1024];
	
	va_start(argptr, fmt);
	vsprintf(text, fmt, argptr);
	va_end(argptr);
	
	ri.Con_Printf(PRINT_ALL, "%s", text);
}

#endif
